# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC
from typing import Any, Callable, Optional, TypeVar

from .model import Model
from ..core import DataObject
from ..core.core_converter import Converter
from ..core.factory import DataObjectFactory


TypeFunction = Callable[[Model], str]
"""Takes a Model instance, and returns the corresponding DataObject type."""


In = TypeVar('In')
"""The input representation type"""


Out = TypeVar('Out')
"""The output representation type"""


class ModelConverter(Converter[Model, DataObject], ABC):
    """
    Converts Sqlalchemy model instances to DataObject instances.
    """


class DefaultModelConverter(ModelConverter):

    def __init__(
        self,
        type_function: TypeFunction,
        data_object_factory: DataObjectFactory,
        max_depth: int = 1,
        requested_fields: list[str] | None = None,
    ) -> None:
        """
        Takes a type_function Callable, which determines the type of the
        DataObject for a given Model instance.
        """

        self.__type_function = type_function
        self.__data_object_factory = data_object_factory

        self.__requested_fields = requested_fields
        self.__max_depth = (
            None
            if self.__requested_fields
            else max_depth
        )

    def convert(self, model: Model) -> DataObject:
        return self.__convert_to_max_depth(
            model,
            self.__initial_marker
        )

    def __convert_to_max_depth(
        self,
        model: Model | None,
        marker: int
    ) -> DataObject:

        if model is None:
            return None

        type_ = self.__type_function(model)

        return self.__data_object_factory(
            type_,
            id_=model.instance_id,
            attributes=model.instance_attributes,
            to_one=self.__convert_to_ones(
                model,
                marker
            )
        )

    @property
    def __initial_marker(self) -> int | str:
        return (
            ''
            if self.__requested_fields
            else 0
        )

    def __convert_to_ones(
        self,
        model: Model,
        marker: int | str
    ) -> dict[str, Optional[DataObject]]:

        if self.__max_depth and marker >= self.__max_depth:
            return {}

        return {
            k: self.__convert_to_max_depth(
                model.instance_to_one_relations[k],
                self.__get_next_marker(
                    k,
                    marker,
                )
            )
            for k in self.__get_requested_to_ones(
                model,
                marker,
            )
        }

    def __get_next_marker(
        self,
        k: str,
        marker: int | str
    ) -> int | str:

        if self.__requested_fields:
            return f'{marker}.{k}' if marker else k
        else:
            return marker + 1

    def __get_requested_to_ones(
        self,
        model_instance: Model,
        marker: int | str
    ) -> list[str]:

        all_keys = list(
            model_instance.get_to_one_relationship_config().keys()
        )

        if not self.__requested_fields:
            return all_keys

        return [
            k for k in all_keys
            if self.__requested_to_one(k, marker)
        ]

    def __requested_to_one(
        self,
        k: str,
        marker: int | str
    ) -> bool:

        next_marker = self.__get_next_marker(
            k,
            marker,
        )

        return any(
            r.startswith(next_marker)
            for r in self.__requested_fields
        )


class DataObjectConverter(Converter[DataObject, Model], ABC):
    """
    Converts `DataObject` instances to `Model` instances.
    """


class DefaultDataObjectConverter(DataObjectConverter):

    def __init__(
        self,
        type_models_dict: dict[str, type[Model]]
    ) -> None:
        """
        `type_models_dict` maps object type to the
        corresponding `type[Model]` class.
        """

        self.__models_dict = type_models_dict

    def convert(self, input_: DataObject) -> Model:
        model_class = self.__models_dict[input_.type]

        return model_class(
            **self.__get_id_dict(input_.id, model_class),
            **input_.attributes,
            **self.__get_relation_dict(
                model_class,
                input_._to_one_objects
            )
        )

    def __get_id_dict(
        self,
        id_: str,
        model_class: type[Model]
    ) -> dict[str, str]:

        id_column_name = model_class.get_id_column_name()
        return {id_column_name: id_}

    def __get_relation_dict(
        self,
        model_class: type[Model],
        ones: dict[str, DataObject]
    ) -> dict[str, str]:
        # TODO validation - relationship names and their types

        return {
            model_class.get_foreign_key_name(
                rel_name
            ): self.__map_to_foreign_key(rel_obj)
            for rel_name, rel_obj in ones.items()
        }

    def __map_to_foreign_key(
        self,
        rel_obj: DataObject | None
    ) -> Any | None:

        return (
            None if rel_obj is None else rel_obj.id
        )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from typing import Iterable

from more_itertools import flatten

from .data_object import DataObject
from .factory import DataObjectFactory
from .operator.updater import DataObjectUpdate


class DataObjectToDataObjectOrUpdateConverter(ABC):
    """
    This is (currently) not inheriting from Converter as we are changing
    the method signatures (convert() can return None, a DataObject or an
    Iterable of DataObjects)
    """

    def __init__(
        self,
        data_object_factory: DataObjectFactory
    ) -> None:
        """
        Takes a data_object_factory to use for creating new DataObjects
        """
        self._data_object_factory = data_object_factory
        self._return_objects = []

    def convert_iterable(
        self,
        inputs: Iterable[DataObject | DataObjectUpdate]
    ) -> Iterable[DataObject]:
        return flatten((self.convert(i) for i in inputs))

    def get_return_objects(self):
        return self._return_objects

    @abstractmethod
    def convert(self, input_: DataObject) -> Iterable[DataObject | DataObjectUpdate]:
        """
        Converts an input representation to an output representation.
        output can be:
            None: ignored
            DataObject
            Iterable of DataObjects
        """


class DefaultDataObjectToDataObjectConverter(DataObjectToDataObjectOrUpdateConverter):

    def __init__(
        self,
        data_object_factory: DataObjectFactory,
        destination_object_type: str | None = None
    ):

        super().__init__(data_object_factory)
        self.__dest_object_type = destination_object_type

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        """
        A "passthrough" converter only dealing with attributes:
            None: ignored
            DataObject
            Iterable of DataObjects
        """
        if data_object is not None and data_object.id is not None:
            dest_type = (
                self.__dest_object_type
                if self.__dest_object_type is not None
                else self.data_loader._destination_object_type
            )
            ret = self._data_object_factory(
                id_=data_object.id,
                type_=dest_type,
                attributes={**data_object.attributes}
            )
            return iter([ret])
        return iter([])



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_specimen_status_to_data_object(status_data, target_datasource):
    CoreDataObject = target_datasource.data_object_factory  # noqa N806
    for row in status_data:
        ret = CoreDataObject('specimen', attributes={
            'tolid': row.get('sample'),
            'specimen_id': row.get('accession'),
            'status_summary': row.get('statussummary'),
            'status': row.get('status')
        })
        yield ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_tolqc_run_data_to_data_object(run_data, target_datasource):
    CoreDataObject = target_datasource.data_object_factory  # noqa N806
    for run_datum in run_data:
        ret = CoreDataObject(
            'run_data',
            attributes={'run_id': run_datum.runs.name,
                        'position': run_datum.runs.element,
                        'tag_index': run_datum.tag_index,
                        'sample_ref': run_datum.samples.name,
                        'public_name': run_datum.samples.specimens.name,
                        'common_name': run_datum.samples.specimens.species.name,
                        'supplier_name': run_datum.samples.specimens.supplied_name,
                        'biospecimen_id': run_datum.samples.specimens.accessions.name,
                        'taxon_id': run_datum.samples.specimens.species.taxon_id,
                        'instrument_model': run_datum.runs.platforms.model,
                        'start_date': run_datum.runs.start_date,
                        'qc_date': run_datum.runs.qc_date,
                        'pipeline_id_lims': '',  # TODO
                        'tag_sequence': run_datum.tag1_id,
                        'tag2_sequence': run_datum.tag2_id,
                        'complete_date': run_datum.runs.date,
                        'manual_qc': run_datum.lims_qc}
        )
        yield ret



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Iterable

from .data_object import DataObject, ErrorObject


class ValidationSeverity(str, Enum):
    ERROR = 'error'
    WARNING = 'warning'


@dataclass(frozen=True, kw_only=True)
class ValidationResult:
    object_id: str
    """Maps to the row number from the original spreadsheet"""

    detail: str
    """A helpful error message"""

    severity: ValidationSeverity
    """Either `'warning'` or `'error'`"""

    field: str | list[str] | None = None
    """
    The field(s) that failed validation

    `None` indicates the all fields failed on this row
    """

    code: str | None = None
    """An (optional) reference to an external error code"""


class Validator(ABC):

    def __init__(self) -> None:
        super().__init__()

        self.__results: list[ValidationResult] = []

    @abstractmethod
    def _validate_data_object(
        self,
        obj: DataObject
    ) -> DataObject:
        """Validates a single `DataObject` instance."""

    def validate(
        self,
        object_stream: Iterable[DataObject | ErrorObject]
    ) -> Iterable[DataObject | ErrorObject]:
        """
        Validates a stream of `DataObject` instances.
        """

        for obj in object_stream:
            if isinstance(obj, ErrorObject):
                yield obj
            else:
                yield self._validate_data_object(obj)

    def add_warning(
        self,
        *,
        object_id: str,
        detail: str,
        field: str | list[str] | None = None,
        code: str | None = None,
    ) -> None:

        self._add_result(
            object_id=object_id,
            detail=detail,
            severity=ValidationSeverity.WARNING,
            field=field,
            code=code,
        )

    def add_error(
        self,
        *,
        object_id: str,
        detail: str,
        field: str | list[str] | None = None,
        code: str | None = None,
    ) -> None:

        self._add_result(
            object_id=object_id,
            detail=detail,
            severity=ValidationSeverity.ERROR,
            field=field,
            code=code,
        )

    @property
    def results(self) -> list[ValidationResult]:
        return self.__results

    @property
    def no_errors(self) -> bool:
        """
        Returns `True` if there are no validation errors.
        """

        return not any(
            r.severity == ValidationSeverity.ERROR
            for r in self.__results
        )

    def _add_result(
        self,
        *,
        object_id: str,
        detail: str,
        severity: ValidationSeverity,
        field: str | list[str] | None = None,
        code: str | None = None,
    ) -> None:

        result = ValidationResult(
            object_id=object_id,
            detail=detail,
            severity=severity,
            field=field,
            code=code,
        )

        self.__results.append(result)



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Iterable as IterableABC
from dataclasses import dataclass
from typing import Iterator

from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import (
    Mapped,
    mapped_column
)

from ..model import Model


@dataclass(frozen=True, kw_only=True)
class SummaryModels(IterableABC[type[Model]]):
    """
    Contains the needed models for summarys.

    """

    summary: type[Model]

    def __iter__(self) -> Iterator[type[Model]]:
        """
        Returns in order they should be deleted
        """

        return iter(
            [
                self.summary
            ]
        )


def create_summary_models(
    base_model_class: type[Model]
) -> SummaryModels:
    """
    Creates all needed models (and joining tables) for
    summaries.

    Returns a `SummaryModels` instance that functions like an
    `Iterable`.
    """

    class Summary(base_model_class):
        __tablename__ = 'summary'

        id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)  # noqa A003

        source_object_type: Mapped[str] = mapped_column(nullable=False)
        destination_object_type: Mapped[str] = mapped_column(nullable=True)

        object_filters: Mapped[dict] = mapped_column(
            JSONB,
            nullable=False,
            default={}
        )

        group_by: Mapped[dict] = mapped_column(
            JSONB,
            nullable=False,
            default=[]
        )

        stats_fields: Mapped[dict] = mapped_column(
            JSONB,
            nullable=False,
            default=[]
        )

        stats: Mapped[dict] = mapped_column(
            JSONB,
            nullable=False,
            default=[]
        )

        prefix: Mapped[str] = mapped_column(
            nullable=False,
            default=''
        )

    return SummaryModels(
        summary=Summary,
    )



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_bold_samples_to_data_object(samples, target_datasource):
    CoreDataObject = target_datasource.data_object_factory  # noqa N806
    mapping = {
        'phylum': 'taxonomy_phylum_taxon_name',
        'class': 'taxonomy_class_taxon_name',
        'order': 'taxonomy_order_taxon_name',
        'family': 'taxonomy_family_taxon_name',
        'subfamily': 'taxonomy_subfamily_taxon_name',
        'genus': 'taxonomy_genus_taxon_name',
        'specimen_id': 'specimen_identifiers_sampleid',
        'collector_sample_id': 'specimen_identifiers_catalognum',
        'institution_storing': 'specimen_identifiers_institution_storing',
        'voucher_status': 'specimen_desc_voucher_status',
        'lifestage': 'specimen_desc_lifestage',
        'organism_part': 'specimen_desc_organism_part',
        'gal': 'collection_event_collectors',
        'collection_country': 'collection_event_country',
        'exact_site': 'collection_event_exactsite',
        'decimal_latitude': 'collection_event_coordinates_lat',
        'decimal_longitude': 'collection_event_coordinates_long',
        'site_code': 'collection_event_site_code',
        'collection_method': 'collection_event_site_sampling_protocol'
    }
    for s in samples:
        attributes = {k: getattr(s, v)
                      for k, v in mapping.items()
                      if v in s.attributes}
        ret = CoreDataObject('sample', attributes)
        yield ret



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from datetime import datetime, timedelta
from typing import Any, Callable, Optional
from urllib.parse import urlencode

import requests
from requests.auth import HTTPBasicAuth

from .models import (
    ModelClass,
    ModelTuple,
    create_models
)
from ..session import (
    SessionFactory,
    create_session_factory
)
from ...api_base.auth import (
    AuthBlueprint,
    AuthManager,
    OidcConfig,
    StateNotFoundError,
    env_oidc_config
)
from ...api_base.misc import AuthContext


class DbAuthManager(AuthManager):
    """
    Implements `AuthManager` using SQLALchemy
    and a database, to achieve an OIDC workflow.
    """

    def __init__(
        self,
        oidc_config: OidcConfig,
        session_factory: SessionFactory,
        model_tuple: ModelTuple,
        state_delete_delta: timedelta,
        oidc_id_target: str,
        oidc_ext_mapping: dict[str, str],
    ) -> None:

        self.__config = oidc_config
        self.__session_factory = session_factory
        self.__models = model_tuple
        self.__state_delta = state_delete_delta
        self.__oidc_id_target = oidc_id_target
        self.__ext_map = oidc_ext_mapping

    def login(self) -> dict[str, str]:
        self.__cleanup_before_login()
        state_uuid = self.__create_new_state()

        return self.__format_login_url(state_uuid)

    def get_token_from_callback(
        self,
        state: str,
        code: str
    ) -> dict[str, str]:

        self.__check_state_exists(state)

        return self.__post_to_token_url(code)

    def create_user_profile(
        self,
        token: str
    ) -> dict[str, Any]:

        oidc_id, oidc_ext = self.__get_oidc_details(token)
        user_id, user_ext = self.__get_or_create_user(
            oidc_id,
            oidc_ext
        )
        token_extra = self.__get_or_create_token(
            token,
            user_id
        )

        return {
            **token_extra,
            **user_ext,
            'oidc_id': oidc_id
        }

    def authenticate(
        self,
        ctx: AuthContext,
        token: str,
    ) -> None:

        user_id, role_names = self.__get_user_details_for_token(token)
        if user_id is not None:
            ctx.user_id = user_id
            ctx.roles = role_names

    def revoke_token(self, token: str) -> None:
        self.__delete_token(token)
        self.__post_revoke(token)

    def __map_oidc_extra(
        self,
        json_return: dict[str, Any]
    ) -> dict[str, Any]:

        return {
            v: json_return.get(k)
            for k, v in self.__ext_map.items()
        }

    def __post_revoke(self, token: str) -> None:
        r = requests.post(
            self.__config.revoke_url,
            data={
                'token': token,
                'client_id': self.__config.client_id,
                'client_secret': self.__config.client_secret
            }
        )
        r.raise_for_status()

    def __get_oidc_details(
        self,
        token: str
    ) -> tuple[str, dict[str, Any]]:

        headers = {
            'Authorization': f'Bearer {token}'
        }

        r = requests.get(
            self.__config.user_info_url,
            headers=headers
        )
        r.raise_for_status()

        json_return = r.json()

        return (
            json_return[self.__oidc_id_target],
            self.__map_oidc_extra(json_return)
        )

    def __get_user_details_for_token(
        self,
        token: str
    ) -> tuple[Optional[str], list[str]]:

        token_model = self.__models.token_class

        with self.__session_factory() as sess:
            instance = token_model.get(sess, token)

            if instance is None:
                return None, []
            else:
                user = instance.user
                return str(user.id), user.role_names

    def __get_or_create_user(
        self,
        oidc_id: str,
        oidc_ext: dict[str, Any]
    ) -> tuple[int, dict[str, Any]]:

        user_model = self.__models.user_class

        with self.__session_factory() as sess:
            user = user_model.get_or_create(
                sess,
                oidc_id,
                **oidc_ext
            )
            userinfo_ext = self.__get_user_ext(
                user
            )

            return user.id, userinfo_ext

    def __get_user_ext(
        self,
        user: Any
    ) -> dict[str, Any]:

        if self.__user_model_defines_ext(user):
            return user.get_userinfo_ext()
        else:
            return {}

    def __user_model_defines_ext(
        self,
        user: Any
    ) -> bool:
        """
        Returns `True` if an `get_userinfo_ext()` method is
        defined on the mixin for `User`.
        """

        return (
            hasattr(user, 'get_userinfo_ext')
            and callable(
                user.get_userinfo_ext
            )
        )

    def __get_or_create_token(
        self,
        token: str,
        user_id: int
    ) -> str:

        token_model = self.__models.token_class

        with self.__session_factory() as sess:
            return token_model.get_or_create(
                sess,
                token,
                user_id
            )

    def __delete_token(self, token: str) -> str:
        token_model = self.__models.token_class

        with self.__session_factory() as sess:
            return token_model.delete(
                sess,
                token
            )

    def __check_state_exists(self, state_uuid: str) -> None:
        state_model = self.__models.state_class

        with self.__session_factory() as sess:
            if not state_model.exists(sess, state_uuid):
                raise StateNotFoundError()

    def __post_to_token_url(self, code: str) -> dict[str, str]:
        r = requests.post(
            self.__config.token_url,
            auth=self.__basic_auth(),
            data=self.__token_post_data(code)
        )
        r.raise_for_status()

        return r.json()

    def __basic_auth(self) -> HTTPBasicAuth:
        return HTTPBasicAuth(
            self.__config.client_id,
            self.__config.client_secret
        )

    def __token_post_data(self, code: str) -> dict[str, str]:
        return {
            'grant_type': 'authorization_code',
            'code': code,
            'redirect_uri': self.__config.redirect_uri
        }

    def __delete_old_states(self) -> None:
        before = datetime.now() - self.__state_delta
        state_model = self.__models.state_class

        with self.__session_factory() as sess:
            state_model.delete_old(sess, before)

    def __cleanup_before_login(self) -> None:
        self.__delete_old_states()
        self.__delete_expired_tokens()

    def __delete_expired_tokens(self) -> None:
        token_model = self.__models.token_class

        with self.__session_factory() as sess:
            token_model.delete_expired(sess)

    def __create_new_state(self) -> str:
        state_model = self.__models.state_class

        with self.__session_factory() as sess:
            return state_model.add(sess)

    def __format_login_url(
        self,
        state_uuid: str
    ) -> dict[str, str]:

        encoded = self.__encode_params(state_uuid)
        login_url = f'{self.__config.auth_url}?{encoded}'

        return {'loginUrl': login_url}

    def __encode_params(self, state_uuid: str) -> str:
        params = {
            'client_id': self.__config.client_id,
            'response_type': 'code',
            'state': state_uuid,
            'redirect_uri': self.__config.redirect_uri,
            'scope': 'openid profile email'
        }

        return urlencode(params)


class DbAuthBlueprint(AuthBlueprint):
    """
    Same as `AuthBlueprint`, but stores auth `Model`
    classes too.
    """

    def __init__(
        self,
        auth_manager: AuthManager,
        url_prefix: str,
        models: ModelTuple
    ) -> None:

        super().__init__(auth_manager, url_prefix)

        self.__models = models

    @property
    def models(self) -> ModelTuple:
        """
        Returns a `ModelTuple` containing the
        created `Model` classes for auth.
        """

        return self.__models


def __db_auth_manager(
    model_base: ModelClass,
    db_uri: str,
    oidc_config: OidcConfig,
    user_mixin_class: ModelClass,
    user_model_name: str,
    token_mixin_class: ModelClass,
    token_is_pk: bool,
    role_mixin_class: ModelClass,
    oidc_id_column_name: str,
    state_delete_delta: timedelta,
    token_expiry_delta: timedelta,
    oidc_id_target: str,
    oidc_ext_mapping: dict[str, str],
    prefix_with_name: bool
) -> tuple[DbAuthManager, ModelTuple]:

    session_factory = create_session_factory(db_uri)

    model_tuple = create_models(
        model_base,
        user_model_name,
        oidc_id_column_name,
        user_mixin_class,
        token_mixin_class,
        token_is_pk,
        role_mixin_class,
        token_expiry_delta,
        prefix_with_name
    )

    auth_manager = DbAuthManager(
        oidc_config,
        session_factory,
        model_tuple,
        state_delete_delta,
        oidc_id_target,
        oidc_ext_mapping,
    )

    return auth_manager, model_tuple


def db_auth_blueprint(
    model_base: ModelClass,
    db_uri: str,

    url_prefix: str = '/auth',
    prefix_with_name: bool = False,
    oidc_config_factory: Callable[[], OidcConfig] = env_oidc_config,
    user_mixin_class: ModelClass = object,
    user_model_name: str = 'user',
    token_mixin_class: ModelClass = object,
    token_is_pk: bool = False,
    role_mixin_class: ModelClass = object,
    oidc_id_column_name: str = 'oidc_id',
    state_delete_delta: timedelta = timedelta(hours=1),
    token_expiry_delta: timedelta = timedelta(days=7),
    oidc_id_target: str = 'email',
    oidc_ext_mapping: dict[str, str] = {}
) -> DbAuthBlueprint:
    """
    Creates a flask `Blueprint` for auth using a DB,
    given a suitable `ModelClass` and connection `db_uri`.

    Returns a `DbAuthBlueprint` instance, containing a `ModelTuple`
    as a member (which itself contains the newly created `Model`
    classes).

    By default, OIDC tokens expire after 7 days, and inter-request
    OIDC state objects are only valid for 1 hour.

    An SQL-Alchemy `user_mixin_class` can be provided, to
    augment the generated `User` model with extra columns.
    However - these must all either:

    - give `nullable=True`
    - provide a default value.
    """

    auth_manager, models = __db_auth_manager(
        model_base,
        db_uri,
        oidc_config_factory(),
        user_mixin_class,
        user_model_name,
        token_mixin_class,
        token_is_pk,
        role_mixin_class,
        oidc_id_column_name,
        state_delete_delta,
        token_expiry_delta,
        oidc_id_target,
        oidc_ext_mapping,
        prefix_with_name
    )

    auth_bp = DbAuthBlueprint(
        auth_manager,
        url_prefix,
        models
    )

    return auth_bp



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Iterable as IterableABC
from dataclasses import dataclass
from datetime import datetime
from typing import Iterator

from sqlalchemy import (
    ForeignKey
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import (
    Mapped,
    mapped_column,
    relationship
)

from ..model import Model


@dataclass(frozen=True, kw_only=True)
class LoaderModels(IterableABC[type[Model]]):
    """
    Contains the needed models for loaders.

    """

    data_source_instance: type[Model]
    loader: type[Model]

    def __iter__(self) -> Iterator[type[Model]]:
        """
        Returns in order they should be deleted
        """

        return iter(
            [
                self.loader,
                self.data_source_instance
            ]
        )


def create_loader_models(
    base_model_class: type[Model]
) -> LoaderModels:
    """
    Creates all needed models (and joining tables) for
    loaders.

    Returns a `LoaderModels` instance that functions like an
    `Iterable`.
    """

    class Loader(base_model_class):
        __tablename__ = 'loader'

        id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)  # noqa A003

        source_object_type: Mapped[str] = mapped_column(nullable=False)
        destination_object_type: Mapped[str] = mapped_column(nullable=False)

        object_filters: Mapped[dict] = mapped_column(
            JSONB,
            nullable=True
        )

        prefix: Mapped[str] = mapped_column(nullable=False, default='')
        convert_class: Mapped[str] = mapped_column(nullable=True)
        candidate_key: Mapped[dict] = mapped_column(JSONB, nullable=True)
        date_last_run: Mapped[datetime] = mapped_column(nullable=True)

        # Frequency of loading
        frequency_weekly: Mapped[bool] = mapped_column(nullable=True)
        frequency_daily: Mapped[bool] = mapped_column(nullable=True)
        frequency_hourly: Mapped[bool] = mapped_column(nullable=True)
        frequency_quarter_hourly: Mapped[bool] = mapped_column(nullable=True)

        # For loading by IDs
        ids_object_type: Mapped[str] = mapped_column(nullable=True)
        ids_attribute: Mapped[str] = mapped_column(nullable=True)
        ids_object_filters: Mapped[dict] = mapped_column(JSONB, nullable=True)
        ids_sort_by: Mapped[str] = mapped_column(nullable=True)
        ids_attribute_in_source: Mapped[str] = mapped_column(nullable=True)

        # Rerlationships
        source_data_source_instance_id: Mapped[int] = mapped_column(
            ForeignKey('data_source_instance.id'),
            nullable=False
        )
        source_data_source_instance: Mapped['DataSourceInstance'] = relationship(  # noqa F821
            back_populates='source_loaders',
            foreign_keys=[source_data_source_instance_id]
        )

        destination_data_source_instance_id: Mapped[int] = mapped_column(
            ForeignKey('data_source_instance.id'),
            nullable=False
        )
        destination_data_source_instance: Mapped['DataSourceInstance'] = relationship(  # noqa F821
            back_populates='destination_loaders',
            foreign_keys=[destination_data_source_instance_id]
        )
        ids_data_source_instance_id: Mapped[int] = mapped_column(
            ForeignKey('data_source_instance.id'),
            nullable=True
        )
        ids_data_source_instance: Mapped['DataSourceInstance'] = relationship(  # noqa F821
            back_populates='ids_loaders',
            foreign_keys=[ids_data_source_instance_id]
        )

    class DataSourceInstance(base_model_class):
        __tablename__ = 'data_source_instance'

        id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)  # noqa A003

        name: Mapped[str] = mapped_column(nullable=False)
        builtin_name: Mapped[str] = mapped_column(nullable=False)

        source_loaders: Mapped[list['Loader']] = relationship(  # noqa F821
            back_populates='source_data_source_instance',
            foreign_keys=[Loader.source_data_source_instance_id]
        )
        destination_loaders: Mapped[list['Loader']] = relationship(  # noqa F821
            back_populates='destination_data_source_instance',
            foreign_keys=[Loader.destination_data_source_instance_id]
        )
        ids_loaders: Mapped[list['Loader']] = relationship(  # noqa F821
            back_populates='ids_data_source_instance',
            foreign_keys=[Loader.ids_data_source_instance_id]
        )

    return LoaderModels(
        loader=Loader,
        data_source_instance=DataSourceInstance
    )



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_species_lab_work_status_to_data_object(species_lab_work_statuses, target_datasource):
    CoreDataObject = target_datasource.data_object_factory  # noqa N806
    for species_lab_work_status in species_lab_work_statuses:
        attribute_name = species_lab_work_status.status.lower() + '_date'
        ret = CoreDataObject(
            'species',
            id_=species_lab_work_status.species.id,
            attributes={attribute_name: species_lab_work_status.updated_at}
        )
        yield ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class BenchlingTissuePrepToElasticTissuePrepConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        ret = self._data_object_factory(
            'tissue_prep',
            data_object.eln_tissue_prep_id,
            attributes={
                'sample': {'id': str(data_object.sts_id)},
                'species': {'id': str(data_object.taxon_id)},
                'tolid': {'id': data_object.programme_id},
                **{k: v
                   for k, v in data_object.attributes.items()
                   if k not in ['eln_tissue_prep_id',
                                'sts_id',
                                'taxon_id',
                                'programme_id']}
            }
        )
        return iter([ret])



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from collections.abc import Mapping
from typing import Callable, Iterator, Optional

from .api_datasource import (
    ApiDataSource,
    DOConverterFactory,
    JsonConverterFactory
)
from .client import JsonApiClient
from .converter import (
    DataObjectConverter,
    JsonApiConverter
)
from .filter import DefaultApiFilter
from .parser import DefaultParser
from .view import DefaultView
from ..core import DataSource


class _ApiDSDict(Mapping):
    def __init__(self, api_ds: ApiDataSource) -> None:
        self.__ds = api_ds

    def __getitem__(self, __k: str) -> ApiDataSource:
        if __k not in self.__ds.supported_types:
            raise KeyError()
        return self.__ds

    def __iter__(self) -> Iterator[str]:
        return iter(self.__ds.supported_types)

    def __len__(self) -> int:
        return len(self.__ds.supported_types)


class _ConverterFactory:
    """
    Manges the instantation of:

    - `DataObjectConverter`
    - `JsonApiConverter`
    """

    def __init__(self, prefix: str) -> None:
        self.__prefix = prefix
        self.__data_source: Optional[DataSource] = None

    @property
    def data_source(self) -> Optional[DataSource]:
        return self.__data_source

    @data_source.setter
    def data_source(
        self,
        ds: DataSource
    ) -> None:

        self.__data_source = ds

    def do_converter_factory(self) -> DOConverterFactory:
        """
        Returns an instantiated `DataObjectConverter`.
        """

        view = DefaultView(prefix=self.__prefix)
        return DataObjectConverter(view)

    def json_converter_factory(self) -> JsonConverterFactory:
        """
        Returns an instantiated `JsonApiConverter`.
        """

        parser = DefaultParser(self.__ds_dict)
        return JsonApiConverter(parser)

    @property
    def __ds_dict(self) -> dict[str, DataSource]:
        return _ApiDSDict(self.data_source)


def _get_client_factory(
    api_url: str,
    token: Optional[str],
    data_prefix: str,
    retries: int
) -> Callable[[], JsonApiClient]:
    """
    A resonable default for creating
    a `JsonApiClient` instance
    """

    return lambda: JsonApiClient(
        api_url,
        token=token,
        data_prefix=data_prefix,
        retries=retries
    )


def _filter_factory() -> DefaultApiFilter:
    return DefaultApiFilter()


def create_api_datasource(
    api_url: str,
    token: Optional[str] = None,

    data_prefix: str = '/data',
    retries: int = 5
) -> ApiDataSource:
    """
    Instantiates `ApiDataSource` using the given:

    - `api_url`
    - `token` (optional)
    """

    client_factory = _get_client_factory(
        api_url,
        token=token,
        data_prefix=data_prefix,
        retries=retries
    )
    manager = _ConverterFactory(data_prefix)

    api_ds = ApiDataSource(
        client_factory,
        manager.json_converter_factory,
        manager.do_converter_factory,
        _filter_factory
    )

    manager.data_source = api_ds

    return api_ds



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT


def convert_portaldb_sequencing_request_events_to_sequencing_requests(sres, target_datasource):
    for sre in sres:
        CoreDataObject = target_datasource.data_object_factory  # noqa N806
        ret = CoreDataObject(
            'sequencing_request',
            attributes={'sample_ref': sre.id, **sre.attributes}
        )
        yield ret



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from collections import defaultdict
from functools import reduce, wraps
from itertools import chain
from typing import (
    Callable,
    Iterable,
    Optional,
    Protocol,
    Union,
)


from .asserts import AuthInspector
from .error import ForbiddenError
from ..misc import (
    AuthContext,
    CtxGetter,
    default_ctx_getter
)
from ...core.datasource_filter import AndFilter
from ...core.operator import OperatorMethod


class InspectorHook(Protocol):
    """
    Decorated by one of the methods of
    `CompositeAuthInspector`
    """

    def __call__(
        self,
        object_type: str,
        op: OperatorMethod,
        auth_ctx: Optional[AuthContext] = None
    ) -> Optional[AndFilter]:
        """
        Args:

        - `str`            - the object_type
        - `OperatorMethod` - the requested operation

        Keyword Args:

        - auth_ctx - is either `None` (if unauthenticated)
                     or `AuthContext` (if authenticated).
                     Suffix signature with `**kwargs` if
                     not needed.

        Returns either:

        - `AndFilter` - if extra filter restrictions
                        are needed
        - `None`      - if not
        """


_TypeHandlerDict = dict[
    str,
    list[InspectorHook]
]


_HookDecorator = Union[
    InspectorHook,
    Callable[[InspectorHook], InspectorHook]
]


class CompositeAuthInspector(AuthInspector):
    """
    Composes many inspection hooks into
    a single `AuthInspector`.
    """

    def __init__(
        self,
        admin_role: str = 'admin',
        ctx_getter: CtxGetter = default_ctx_getter
    ) -> None:

        self.__admin_role = admin_role
        self.__ctx_getter = ctx_getter

        self.__noauths: list[InspectorHook] = []
        self.__auths: list[InspectorHook] = []
        self.__typed_noauths = self.__new_type_handler_dict()
        self.__typed_auths = self.__new_type_handler_dict()

    def __call__(
        self,
        object_type: str,
        method: OperatorMethod
    ) -> Optional[AndFilter]:

        ctx = self.__ctx_getter()

        if ctx.authenticated is True:
            if self.__admin_role in ctx.roles:
                return
            else:
                return self.__invoke_auth(
                    object_type,
                    method,
                    ctx
                )
        else:
            return self.__invoke_noauth(
                object_type,
                method
            )

    def noauth(
        self,
        hook: Optional[InspectorHook] = None,
        *,
        object_type: str | list[str] | None = None
    ) -> _HookDecorator:
        """
        Registers a hook `Callable` for a
        request for which no user has
        authenticated.

        Specify an indvidual `str` or `list[str]`
        (`object_type`) to limit its invocation,
        otherwise it will be invoked for all
        types.
        """

        return self.__hook_decorator(
            hook,
            lambda h: self.__noauth_append(
                h,
                object_type
            )
        )

    def auth(
        self,
        hook: Optional[InspectorHook] = None,
        *,
        object_type: str | list[str] | None = None
    ) -> _HookDecorator:
        """
        Registers a hook `Callable` for an
        authenticated user.

        Specify an indvidual `str` or `list[str]`
        (`object_type`) to limit its invocation,
        otherwise it will be invoked for all
        types.
        """

        return self.__hook_decorator(
            hook,
            lambda h: self.__auth_append(
                h,
                object_type
            )
        )

    def always(
        self,
        hook: Optional[InspectorHook] = None,
        *,
        object_type: str | list[str] | None = None
    ) -> _HookDecorator:
        """
        Registers a hook `Callable` that is always
        invoked, indepenently of the authentication
        status of the request.

        Specify an indvidual `str` or `list[str]`
        (`object_type`) to limit its invocation,
        otherwise it will be invoked for all
        types.
        """

        return self.__hook_decorator(
            hook,
            lambda h: self.__always_append(
                h,
                object_type
            )
        )

    def forbid(
        self,
        object_type: str | list[str]
    ) -> None:
        """
        Always forbids any and all operations, to every
        non-admin request, on the given `object_type`(s).
        """

        @self.always(object_type=object_type)
        def __hook(*args, **kwargs) -> None:
            raise ForbiddenError

    def forbid_noauth(
        self,
        object_type: str | list[str]
    ) -> None:
        """
        Forbids any and all operations, to every
        unauthenticated request, on the given
        `object_type`(s).
        """

        @self.noauth(object_type=object_type)
        def __hook(*args, **kwargs) -> None:
            raise ForbiddenError

    def __noauth_append(
        self,
        hook: InspectorHook,
        object_type: str | list[str] | None,
    ) -> None:

        if object_type is None:
            self.__noauths.append(hook)
        else:
            self.__append_to_dict(
                hook,
                object_type,
                self.__typed_noauths
            )

    def __auth_append(
        self,
        hook: InspectorHook,
        object_type: str | list[str] | None,
    ) -> None:

        if object_type is None:
            self.__auths.append(hook)
        else:
            self.__append_to_dict(
                hook,
                object_type,
                self.__typed_auths
            )

    def __always_append(
        self,
        hook: InspectorHook,
        object_type: str | list[str] | None,
    ) -> None:

        self.__noauth_append(hook, object_type)
        self.__auth_append(hook, object_type)

    def __hook_decorator(
        self,
        hook: InspectorHook | None,
        append_func: Callable[[InspectorHook], None]
    ) -> InspectorHook:

        def decorator(
            arg_hook: InspectorHook
        ) -> _HookDecorator:

            append_func(arg_hook)

            @wraps(arg_hook)
            def wrapper(
                __type: str,
                __op: OperatorMethod,
                auth_ctx: AuthContext | None = None
            ):

                return arg_hook(
                    __type,
                    __op,
                    auth_ctx=auth_ctx
                )

            return wrapper

        if callable(hook):
            return decorator(hook)
        else:
            return decorator

    def __append_to_dict(
        self,
        hook: InspectorHook,
        object_type: str | list[str],
        target: dict[str, list[InspectorHook]]
    ) -> None:

        def __append_single(__type: str) -> None:
            existing = target.get(__type, [])
            existing.append(hook)
            target[__type] = existing

        if isinstance(object_type, str):
            __append_single(object_type)
        else:
            for __type in object_type:
                __append_single(__type)

    def __accumulate(
        self,
        existing: Optional[AndFilter],
        add: Optional[AndFilter]
    ) -> Optional[AndFilter]:

        if add is None:
            return existing
        else:
            if existing is None:
                return add
            else:
                return existing | add

    def __invoke_noauth(
        self,
        object_type: str,
        op: OperatorMethod
    ) -> Optional[AndFilter]:

        hooks = self.__get_noauth_hooks(
            object_type
        )

        return reduce(
            lambda d, h: self.__accumulate(
                d,
                h(object_type, op)
            ),
            hooks,
            None
        )

    def __invoke_auth(
        self,
        object_type: str,
        op: OperatorMethod,
        auth_context: AuthContext
    ) -> Optional[AndFilter]:

        hooks = self.__get_auth_hooks(
            object_type
        )

        return reduce(
            lambda d, h: self.__accumulate(
                d,
                h(
                    object_type,
                    op,
                    auth_context=auth_context
                )
            ),
            hooks,
            None
        )

    def __new_type_handler_dict(
        self
    ) -> _TypeHandlerDict:

        return defaultdict(
            lambda: []
        )

    def __get_auth_hooks(
        self,
        object_type: str
    ) -> Iterable[InspectorHook]:

        return chain(
            self.__auths,
            self.__typed_auths[object_type]
        )

    def __get_noauth_hooks(
        self,
        object_type: str
    ) -> Iterable[InspectorHook]:

        return chain(
            self.__noauths,
            self.__typed_noauths[object_type]
        )
